#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cmath>
#include <cstdio>

// Optimized CUDA kernel for convolution, min-reduction over channels and double tanh activation.
// This kernel combines the best features from two implementations:
// - Uses __ldg for read-only memory accesses
// - Manually unrolls the inner loops for the common kernel size of 3
// - Applies #pragma unroll for generic kernel sizes to aid compiler optimizations
// - Uses a grid-stride loop to efficiently cover all output pixels

__global__ void conv_min_tanh_forward_kernel_optimized(
    const float* __restrict__ x,      // Input: [B, C_in, H, W]
    const float* __restrict__ weight, // Weight: [C_out, C_in, K, K]
    const float* __restrict__ bias,   // Bias: [C_out]
    float* __restrict__ output,       // Output: [B, 1, H_out, W_out]
    const int batch,
    const int in_channels,
    const int in_height,
    const int in_width,
    const int out_channels,
    const int kernel_size,
    const int out_height,
    const int out_width) {

    int tid = blockIdx.x * blockDim.x + threadIdx.x;
    int num_pixels = batch * out_height * out_width;
    
    // Grid-stride loop to cover all output pixels
    while (tid < num_pixels) {
        // Map tid to (b, out_y, out_x)
        int b = tid / (out_height * out_width);
        int rem = tid % (out_height * out_width);
        int out_y = rem / out_width;
        int out_x = rem % out_width;

        float min_val = 1e20f;

        // Iterate over all output channels
        for (int oc = 0; oc < out_channels; oc++) {
            // Load bias from read-only cache
            float conv_sum = __ldg(&bias[oc]);

            // Process all input channels
            for (int ic = 0; ic < in_channels; ic++) {
                int base_input = b * (in_channels * in_height * in_width) + ic * (in_height * in_width);
                int base_weight = oc * (in_channels * kernel_size * kernel_size) + ic * (kernel_size * kernel_size);

                if (kernel_size == 3) {
                    // Manually unrolled convolution for kernel size 3 to reduce loop overhead
                    int in_y = out_y;
                    int in_x = out_x;
                    conv_sum += __ldg(&x[base_input + (in_y + 0) * in_width + (in_x + 0)]) * __ldg(&weight[base_weight + 0]);
                    conv_sum += __ldg(&x[base_input + (in_y + 0) * in_width + (in_x + 1)]) * __ldg(&weight[base_weight + 1]);
                    conv_sum += __ldg(&x[base_input + (in_y + 0) * in_width + (in_x + 2)]) * __ldg(&weight[base_weight + 2]);
                    conv_sum += __ldg(&x[base_input + (in_y + 1) * in_width + (in_x + 0)]) * __ldg(&weight[base_weight + 3]);
                    conv_sum += __ldg(&x[base_input + (in_y + 1) * in_width + (in_x + 1)]) * __ldg(&weight[base_weight + 4]);
                    conv_sum += __ldg(&x[base_input + (in_y + 1) * in_width + (in_x + 2)]) * __ldg(&weight[base_weight + 5]);
                    conv_sum += __ldg(&x[base_input + (in_y + 2) * in_width + (in_x + 0)]) * __ldg(&weight[base_weight + 6]);
                    conv_sum += __ldg(&x[base_input + (in_y + 2) * in_width + (in_x + 1)]) * __ldg(&weight[base_weight + 7]);
                    conv_sum += __ldg(&x[base_input + (in_y + 2) * in_width + (in_x + 2)]) * __ldg(&weight[base_weight + 8]);
                } else {
                    // Use unrolling hints for generic kernel sizes
                    #pragma unroll
                    for (int ky = 0; ky < kernel_size; ky++) {
                        #pragma unroll
                        for (int kx = 0; kx < kernel_size; kx++) {
                            int in_y = out_y + ky;
                            int in_x = out_x + kx;
                            int x_index = base_input + in_y * in_width + in_x;
                            int w_index = base_weight + ky * kernel_size + kx;
                            conv_sum += __ldg(&x[x_index]) * __ldg(&weight[w_index]);
                        }
                    }
                }
            }
            // Update minimum using fminf to help avoid branch divergence
            min_val = fminf(min_val, conv_sum);
        }

        // Apply double tanh activation
        float activated = tanhf(tanhf(min_val));
        output[tid] = activated;

        tid += blockDim.x * gridDim.x;
    }
}

// Launcher function
void conv_min_tanh_forward_cuda(
    at::Tensor x,
    at::Tensor conv_weight,
    at::Tensor conv_bias,
    at::Tensor output) {

    const int batch = x.size(0);
    const int in_channels = x.size(1);
    const int in_height = x.size(2);
    const int in_width = x.size(3);

    const int out_channels = conv_weight.size(0);
    const int kernel_size = conv_weight.size(2);  // Assuming square kernel
    const int out_height = in_height - kernel_size + 1;
    const int out_width = in_width - kernel_size + 1;

    int num_pixels = batch * out_height * out_width;
    const int threads = 256;
    const int blocks = (num_pixels + threads - 1) / threads;

    conv_min_tanh_forward_kernel_optimized<<<blocks, threads>>>(
        x.data_ptr<float>(),
        conv_weight.data_ptr<float>(),
        conv_bias.data_ptr<float>(),
        output.data_ptr<float>(),
        batch,
        in_channels,
        in_height,
        in_width,
        out_channels,
        kernel_size,
        out_height,
        out_width
    );

    cudaError_t err = cudaGetLastError();
    if (err != cudaSuccess) {
        printf("Error in conv_min_tanh_forward_kernel_optimized: %s\n", cudaGetErrorString(err));
    }
}

// C++ interface (exposed to Python via pybind11)
at::Tensor forward(
    at::Tensor x,
    at::Tensor conv_weight,
    at::Tensor conv_bias) {

    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(conv_weight.is_cuda(), "conv_weight must be a CUDA tensor");
    TORCH_CHECK(conv_bias.is_cuda(), "conv_bias must be a CUDA tensor");

    const int in_height = x.size(2);
    const int in_width = x.size(3);
    const int kernel_size = conv_weight.size(2);
    const int out_height = in_height - kernel_size + 1;
    const int out_width = in_width - kernel_size + 1;
    const int batch = x.size(0);

    // Allocate output tensor with shape [batch, 1, out_height, out_width]
    auto output = at::empty({batch, 1, out_height, out_width}, x.options());
    conv_min_tanh_forward_cuda(x, conv_weight, conv_bias, output);
    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Optimized convolution with min reduction and double tanh activation (CUDA)");
}
